#include "defs.h"
#include "forkpty.h"

unsigned char ForkPTY_InterruptCode[7] = { 0x9C,0x9A, 0xCB,0xF4, 0x90,0x90, 0xCB };

void ForkPTY_Init()
{
    if(ForkPTY_InterruptCode[5] != 0x90) return;

    /* See if we can find the INT 0x21 handler in DOSBox ROM.
     * If we can, we call the stub directly.
     * This enables it to work even if we loaded FreeDOS etc.
     */
    const unsigned short _far* intptr = (const unsigned short _far*)MK_FP(0xF000u, 0x1000u);
    for(; FP_OFF(intptr) < 0x1000u+128*32; intptr += 16)
    {
        if(intptr[0] == 0xFEFBu
        && intptr[3] == 0x51CBu
        && intptr[4] == 0x40B9u)
        {
            // Initialize with:
            //   PUSHF
            //   CALL FAR 1234:5678
            //   RETF
            *(const unsigned short _far* _near*)&ForkPTY_InterruptCode[2] = intptr;
            return;
        }
    }
    *(unsigned short _near*)&ForkPTY_InterruptCode[0] = 0x21CD; // INT 21h
    ForkPTY_InterruptCode[5] = 0;
}

void ForkPTY::Open()
{
    ForkPTY_Init();
    //fd=1; return;
    _asm {
      mov ax, 0x5300
      mov cx, Rows
      mov dx, Columns
      call far ptr ForkPTY_InterruptCode
      les bp, this     /* Note: BC will push&pop BP around this function */
      mov es:[bp], bx    /* Save fd */
      mov es:[bp+2], cx  /* Save pid */
    }
}
void ForkPTY::Close()
{
    //fd=-1; return;
    _asm {
        les bx, this
        push bx
         mov ax, 0x5301
         mov cx, es:[bx+2] /* pid */
         mov bx, es:[bx]   /* fd */
         call far ptr ForkPTY_InterruptCode
        pop bx
        mov ax,-1 ; mov es:[bx], ax
    }
}
void ForkPTY::Send(const void _far* buffer, unsigned _far& nwrite, unsigned _far& error)
{
    //error=0; return;
    _asm {
       les bx, nwrite ; mov cx, es:[bx] /* nwrite */
       les bx, this   ; mov bx, es:[bx] /* fd */
       push ds
        mov ax, 0x5302
        lds dx, buffer
        call far ptr ForkPTY_InterruptCode
       pop ds
       les bx, nwrite ; mov es:[bx], cx /* store to nwrite */
       les bx, error  ; mov es:[bx], dx /* store to error */
    }
}
void ForkPTY::Recv(void _far* buffer, unsigned _far& nread, unsigned _far& error)
{
    //nread=0;error=0; return;
    _asm {
       les bx, nread  ; mov cx, es:[bx] /* nread */
       les bx, this   ; mov bx, es:[bx] /* fd    */
       push ds
        mov ax, 0x5303
        lds dx, buffer
        call far ptr ForkPTY_InterruptCode
       pop ds
       les bx, nread  ; mov es:[bx], cx /* store to nread */
       les bx, error  ; mov es:[bx], dx /* store to error */
    }
}
